#include <stdio.h>
#include <malloc.h>
#include "perplex.h"

/* This example will use PerpleX wrapper to evaluate
   the stable phases of given compositions (array 'comp'
   below) in pressure-temperature range from 1 bar/1373 K
   to 35 kbar/2173 K, and output them to output.csv 

   To run this example, do:
     $ make
     $ cd datfiles
     $ ../gridding
   Output will be to output.csv

*/

int main(int argc, char **argv) {
	double P, T;
	int ncomp;
	double *comp;
	int i, j, row;

	FILE *fp;

	double p1, p2, t1, t2;
	double dp, dt;
	
	int nphases;
	double *wtphases = 0;
	double *cphases = 0;
	char *namephases = 0;
	double *sysprop = 0;

	int phaseq_dbg = 0; /* debug output from phaseq()? */
	
	/* only call once per build file */
	ini_phaseq("in26klb_new");	// filename (without .dat extension) 
	                                // generated by PerpleX BUILD.
					// If no path defined, it will be looked
					// for in the current folder where
					// this program is executed.
	
	p1 = 1;
	p2 = 35000;
	t1 = 1373;
	t2 = 2173;

	dp = 250;
	dt = 20;	

	ncomp = 6;

	fp = fopen("output.csv", "w");
	if (fp == 0) {
		fprintf(stderr, "!! Failed to open output.csv\n");
		return -1;
	}
	comp = (double*)calloc(ncomp, sizeof(double));

	fprintf(fp, "\"P\",\"T\",\"phase\",\"phasewt\"");
	for (j = 0; j < ncomp; j++) {
		fprintf(fp, ",\"comp%d\"", j);
	}
	fprintf(fp, "\n");

	row = 0;

	wtphases = (double *)malloc(sizeof(double) * p_size_phases);
	cphases = (double *)malloc(sizeof(double) * p_size_phases * p_size_components);
	sysprop = (double *)malloc(sizeof(double) * p_size_sysprops);
	namephases = (char *)malloc(sizeof(char) * p_size_phases * p_pname_len);

	for (P = p1; P <= p2; P += dp) {
		for (T = t1; T <= t2; T += dt) {
			fprintf(stderr, "\rP = %g   T = %g        ", P, T);
            /* KLB-1 */
			/* Array comp should have the composition in the order
             * given by the wrapper function print_comp_order()
             * (see perplex.c)
             */
                        comp[0] = 0.4448;
                        comp[1] = 0.0359;
                        comp[2] = 0.0810;
                        comp[3] = 0.3922;
                        comp[4] = 0.0344;
                        comp[5] = 0.0030;
	
			phaseq(P, T, ncomp, comp, &nphases, wtphases, cphases, sysprop, namephases, phaseq_dbg);
	
			for (i = 0; i < nphases; i++) {
				row++;
				fprintf(fp, "%g,%g,", P, T);			// P , T ,
				fprintf(fp, "%.14s,", namephases + i*p_pname_len);	// name ,
				fprintf(fp, "%.5e", wtphases[i]);			// wt% ,
				for (j = 0; j < ncomp; j++) {
					fprintf(fp, ",%.5e", cphases[i*ncomp + j]);	// composition
				}
				fprintf(fp, "\n");
			}
		}
	}

	free(comp);
	fclose(fp);

	fprintf(stderr, "\n\n");
	
	return 0;
}

