#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>
#include "perplex.h"    /* defines: p_pname_len (phase name length) */
#include "errormin.h"

int lenAllSolPhaseNames;
int totalSolModelCombinations;
int curCombination;
int *allCombinations;
int *numAllSolPhaseNames;
char *allSolPhaseNames;
char *solPhaseNames;
char *genPhaseNames;
char *compNames;
FILE *fp_results;

// num of components; num of phases; current line read in experiment.dat;
// number of variations (combinations of solution models); unique id for calculated result
int num_c, num_ph, curline, num_var, calc_res_id;

ExpVariations meas;
ExpVariations calc;


void safefree(void **p) {
    if(p != NULL && *p != NULL) {
        free(*p);
        *p = NULL;
    }
    return;
}

void permComb (int row) {
    int i, j;

    if (row == num_ph) {
        curCombination++;
    } else {
        for (i = 0; i < numAllSolPhaseNames[row]; i++) {
            for (j = 0; j < (totalSolModelCombinations - curCombination); j++) {
                allCombinations[(curCombination+j) * num_ph + row] = i;
            }
            permComb(row+1);
        }
    }

    return;
}

void initialize() { 
    FILE *fp, *fp_comb;
    char line[MAX_LINE_LENGTH];
    char *str1, *str2;
    int n, nc, np, nres, nextline, errno;
    int i, ires, iphase, j, temp_comb;
    double *temp = NULL;
    
    num_var = 1;    // number of variations, will be updated later...

    curline = 0;
    calc_res_id = 0;
    
    fp = fopen(FILE_EXPDATA, "r");
    if (fp == 0) {
        fprintf(stderr, "!! Opening file %s failed. Exiting.\n", FILE_EXPDATA);
        exit(1);
    }

    fp_results = fopen("errormin_results.dat", "w");
    if (fp_results == NULL) {
        fprintf(stderr, "!! Opening file %s failed. Exiting.\n", "errormin_results.dat");
        exit(1);
    }

    // read NC, NP, NRES
    errno = 0;
    nextline = 1;
    while(nextline) {
        fgets(line, MAX_LINE_LENGTH, fp);
        curline++;
        if(line[0] == '#') continue;
        n = sscanf(line, "%d %d %d", &nc, &np, &nres);
        if (n == 3) {
            // success
            nextline = 0;
        } else if (n > 0) {
            fprintf(stderr, "!! Parse error while reading %s, line %d\n", FILE_EXPDATA, curline);
            exit(1);
        } else if (errno != 0) {
            perror("scanf");
            exit(1);
        } else  {
            // no matching characters, try again if not EOF
            if (feof(fp)) {
                fprintf(stderr, "!! EOF while parsing %s, line %d\n", FILE_EXPDATA, curline);
                exit(1);
            }
            nextline = 1;
        }
    }
    if (DBGL > 5) fprintf(stderr, "!! Read: %d components, %d phases, %d results\n", nc, np, nres);
	    
    	    
    
    num_c = nc; // num of components
    num_ph = np; // num of phases

    // reserve memory for measured experiments
    meas.experiments = (Experiments *)calloc(1, sizeof(Experiments)); // measured data is one experiment

    meas.experiments[0].count = nres; // number of results in this experiment 
    meas.experiments[0].result = (Result *)calloc(nres, sizeof(Result));

    
    // reserve memory and read component names
    compNames = (char *)calloc(nc * C_NAME_LEN, sizeof(char));
    nextline = 1;
    while(nextline) {
        if(feof(fp)) {
            fprintf(stderr, "!! Unexpected EOF\n");
            exit(1);
        }
        fgets(line, MAX_LINE_LENGTH, fp);
        curline++;
        str1 = (char *)calloc(MAX_LINE_LENGTH, sizeof(char));
        nextline = 0;
        strncpy(str1, line, MAX_LINE_LENGTH);
        for (i = 0; ; i++, safefree((void**)(&str1))) {
            str2 = strtok(str1, " \n\t");
            if (str2 == NULL) break;
            if (str2[0] == '#') {
                safefree((void**)(&str1));
                break;
            }            
            if (i < nc) {
                if(strlen(str2) > C_NAME_LEN-1) {
                    fprintf(stderr, "!! Component name (%s) too long.\n", str2);
                    break;
                }
                strncpy((char *)(compNames + i*C_NAME_LEN), str2, C_NAME_LEN);
            }
            if (DBGL > 10) fprintf(stderr, "!! Token: [%s]\n", str2);
        }
        if (i != nc) {
            if (DBGL > 15) fprintf(stderr, "!! Read %d component names instead of expected %d (empty line?).\n", i, nc);
            nextline = 1;
        }
    }
    
    if (DBGL > 5) {
        for (i = 0; i < nc; i++) {
            fprintf(stderr, "Comp %d: %s\n", i, compNames + i*C_NAME_LEN);
        }
    }


    // reserve memory and read common phase names
    genPhaseNames = (char *)calloc(np * p_pname_len, sizeof(char));
    nextline = 1;
    while(nextline) {
        if(feof(fp)) {
            fprintf(stderr, "!! Unexpected EOF\n");
            exit(1);
        }
        str1 = (char *)calloc(MAX_LINE_LENGTH, sizeof(char));
        fgets(line, MAX_LINE_LENGTH, fp);
        curline++;
        nextline = 0;
        strncpy(str1, line, MAX_LINE_LENGTH);
        for (i = 0; ; i++, safefree((void**)(&str1))) {
            str2 = strtok(str1, " \n");
            if (str2 == NULL) break;
            if (str2[0] == '#') {
                safefree((void**)(&str1));
                break;
            }
            if (i < np) {
                if(strlen(str2) > p_pname_len-1) {
                    fprintf(stderr, "!! Phase name (%s) too long.\n", str2);
                    break;
                }
                strncpy((char *)(genPhaseNames + i*p_pname_len), str2, p_pname_len);
            }
            if (DBGL > 10) fprintf(stderr, "!! Token: [%s]\n", str2);
        }
        if (i != np) {
            if (i > 0) {
                if (DBGL > 5) fprintf(stderr, "!! Read %d phase names instead of expected %d.\n", i, np);
                exit(1);
            }
            nextline = 1;
        }
    }
    
    if (DBGL > 5) {
        for (i = 0; i < np; i++) {
            fprintf(stderr, "Phase %d: %s\n", i, genPhaseNames + i*p_pname_len);
        }
    }

    // variations of solution models
    //
    // number of variations per phase
    totalSolModelCombinations = 1;
    lenAllSolPhaseNames = 0;
    safefree((void **)(&temp));
    temp = (double *)calloc(num_ph, sizeof(double));
    numAllSolPhaseNames = (int *)calloc(num_ph, sizeof(int));
    if (readNumArray(fp, temp, num_ph)) {
        fprintf(stderr, "!! Error reading numbers of variations of solution models.\n");
        exit(1);
    }    
    for (i = 0; i < num_ph; i++) {
        numAllSolPhaseNames[i] = (int)(temp[i]);
        totalSolModelCombinations *= numAllSolPhaseNames[i];
        lenAllSolPhaseNames += numAllSolPhaseNames[i];
    }
    if (DBGL > 10) {
        fprintf(stderr, "Total solution model combinations: %d\n", totalSolModelCombinations);
    }
    allSolPhaseNames = (char *)calloc(lenAllSolPhaseNames * p_pname_len, sizeof(char));

    // read variations
    for (j = 0; j < num_ph; j++) {
        safefree((void **)(&str1));
        safefree((void **)(&str2));
        nextline = 1;
        fprintf(stderr, "Reading solution models for [%s]. Expecting %d models.\n", genPhaseNames + j*p_pname_len,  numAllSolPhaseNames[j]);
        while(nextline) {
            if(feof(fp)) {
                fprintf(stderr, "!! Unexpected EOF\n");
                exit(1);
            }
            str1 = (char *)calloc(MAX_LINE_LENGTH, sizeof(char));
            fgets(line, MAX_LINE_LENGTH, fp);
            curline++;
            nextline = 0;
            strncpy(str1, line, MAX_LINE_LENGTH);
            for (i = 0; ; i++, safefree((void**)(&str1))) {
                str2 = strtok(str1, " \n");
                if (str2 == NULL) break;
                if (str2[0] == '#') {
                    safefree((void**)(&str1));
                    break;
                }
                if (i < numAllSolPhaseNames[j]) {
                    if(strlen(str2) > p_pname_len-1) {
                        fprintf(stderr, "!! Phase name (%s) too long.\n", str2);
                        break;
                    }
                    strncpy((char *)(allSolPhaseNames + irregArrPos(numAllSolPhaseNames, j, i)*p_pname_len), str2, p_pname_len);
                }
                if (DBGL > 10) fprintf(stderr, "!! Token: [%s]\n", str2);
            }
            if (i != numAllSolPhaseNames[j]) {
                if (DBGL > 5) {
                    fprintf(stderr, "!! Read %d phase names instead of expected %d.\n", i, numAllSolPhaseNames[j]);
                }
                nextline = 1;
            }
        }
    }
    for (j = 0; DBGL > 10 && j < lenAllSolPhaseNames; j++) {
        fprintf(stderr, " - %.14s\n", allSolPhaseNames + j*p_pname_len);
    }

    // construct all the combinations of solution models
    curCombination = 0;
    allCombinations = (int *)calloc(totalSolModelCombinations * num_ph, sizeof(int));
    permComb(0);    // recursively forms all the combinations
    if (DBGL > 10) {
        for (i = 0; i < totalSolModelCombinations; i++) {
            fprintf(stderr, "Combination %d:\t", i);
            for (j = 0; j < num_ph; j++) {
                fprintf(stderr, "%d ", allCombinations[num_ph * i + j]);
            }
            fprintf(stderr, "\n");
        }
    }
    curCombination = -1;

    // output combinations and corresponding experiment numbers to file
    fp_comb = fopen("calc_combinations.dat", "w");
    if (fp_comb == NULL) {
        fprintf(stderr, "!! Error opening file calc_combinations.dat for writing\n");
        exit(1);
    }
    for (i = 0; i < totalSolModelCombinations; i++) {
        fprintf(fp_comb, "%d ", i);
        for (j = 0; j < num_ph; j++) {
            temp_comb = allCombinations[i * num_ph + j];
            fprintf(fp_comb, "%s ", allSolPhaseNames + irregArrPos(numAllSolPhaseNames, j, temp_comb)*p_pname_len);
        }
        fprintf(fp_comb, "\n");
    }
    fclose(fp_comb);

    // now that we know the number of combinations (variations), let's reserve memory for results
    num_var = totalSolModelCombinations;
    calc.experiments = (Experiments *)calloc(num_var, sizeof(Experiments)); // calculated may have multiple variations
    for (i = 0; i < num_var; i++) {
        calc.experiments[i].count = nres;
        if (!DYNAMIC_MEM_ALLOC) {
            if (DBGL > 5) fprintf(stderr, "Reserving %ld kilobytes mem for calc results...\n", num_var*nres*sizeof(Result));
            calc.experiments[i].result = (Result *)calloc(nres, sizeof(Result));
        }
    }

    // read the experimental data itself
    for (ires = 0; ires < nres; ires++) {
        if(DBGL > 10) fprintf(stderr, "!! Reading result values %d\n", ires);

        // read id, type, P, T, flag1
        errno = 0;
        nextline = 1;
        while(nextline) {
            fgets(line, MAX_LINE_LENGTH, fp);
            curline++;
            if(line[0] == '#') continue;
            n = sscanf(line, "%d %d %lg %lg %d", &(meas.experiments[0].result[ires].id), 
                    &(meas.experiments[0].result[ires].type), &(meas.experiments[0].result[ires].P), 
                    &(meas.experiments[0].result[ires].T), &(meas.experiments[0].result[ires].flag_ph_ab));
            if (n == 5) {
                // success
                nextline = 0;
            } else if (n > 0) {
                fprintf(stderr, "!! Parse error while reading %s, line %d\n", 
                        FILE_EXPDATA, curline);
                exit(1);
            } else if (errno != 0) {
                perror("scanf");
                exit(1);
            } else  {
                // no matching characters, try again if not EOF
                if (feof(fp)) {
                    fprintf(stderr, "!! EOF while parsing %s, line %d\n", 
                            FILE_EXPDATA, curline);
                    exit(1);
                }
                nextline = 1;
            }
        }
        if (DBGL > 5) fprintf(stderr, 
                          "!! Read: id = %d, type = %d, P = %g, T = %g, flag_ph_ab = %d\n", 
                          meas.experiments[0].result[ires].id, meas.experiments[0].result[ires].type, 
                          meas.experiments[0].result[ires].P, meas.experiments[0].result[ires].T,
                          meas.experiments[0].result[ires].flag_ph_ab);

        meas.experiments[0].result[ires].exp = 0;
        meas.experiments[0].result[ires].id = ires;
                          
        // read bulk composition
        meas.experiments[0].result[ires].bulkcomp = (double *)calloc(num_c, sizeof(double));
        if (readNumArray(fp, meas.experiments[0].result[ires].bulkcomp, num_c)) {
            fprintf(stderr, "!! Error reading bulk composition.\n");
            exit(1);
        }

        // read phase amounts
        meas.experiments[0].result[ires].phasewt = (double *)calloc(num_ph, sizeof(double));
        if (readNumArray(fp, meas.experiments[0].result[ires].phasewt, num_ph)) {
            fprintf(stderr, "!! Error reading phase amounts.\n");
            exit(1);
        }
        
        // read phase compositions
        meas.experiments[0].result[ires].phasecomp = (double *)calloc(num_ph*num_c, sizeof(double));
        for (iphase = 0; iphase < num_ph; iphase++) {
            fprintf(stderr, "!! Reading phase composition for phase %d...\n", iphase);
            if (readNumArray(fp, &(meas.experiments[0].result[ires].phasecomp[iphase*num_c]), num_c)) {
                fprintf(stderr, "!! Error reading phase compositions.\n");
                exit(1);
            }
        }
        
    }
    
    fclose(fp);

    solPhaseNames = (char *)calloc(num_ph * p_pname_len, sizeof(char));
    

    /*strncpy((char *)(solPhaseNames + 0*p_pname_len), "O(HP)         ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 1*p_pname_len), "Opx(HP)       ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 2*p_pname_len), "Cpx(HP)       ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 3*p_pname_len), "Pl(h)         ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 4*p_pname_len), "Gt(HP)        ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 5*p_pname_len), "Sp(HP)        ", p_pname_len-1);
    strncpy((char *)(solPhaseNames + 6*p_pname_len), "pMELTS(G)     ", p_pname_len-1);*/
}

int irregArrPos(int *linelens, int line, int col) {
    int sum = 0;
    int i;

    for (i = 0; i < line; i++) {
        sum = sum + linelens[i];
    }
    sum = sum + col;
    return sum;
}

int nextSolutionModelCombination() {
    FILE *fread, *fwrite;
    int i, solChoice, c, pad;
    char ch;

    if (curCombination >= totalSolModelCombinations - 1) return 0;   // no more combinations
    
    curCombination++;

    fread = fopen("datfiles/build_header.dat", "r");
    fwrite = fopen("datfiles/errormin.dat", "w");
    if (fread == 0 || fwrite == 0) {
        fprintf(stderr, "!! Error opening BUILD configuration file / output file\n");
        exit(1);
    }
    do {
        ch = fgetc(fread);
        if (ch != EOF) fputc(ch, fwrite);
        else break;
    } while(1);

    fclose(fread);

    for (i = 0; i < num_ph; i++) {
        solChoice = allCombinations[curCombination * num_ph + i];
        strncpy((char *)(solPhaseNames + i*p_pname_len), &(allSolPhaseNames[irregArrPos(numAllSolPhaseNames, i, solChoice)*p_pname_len]), p_pname_len-1);

        fprintf(fwrite, "%s\n", (char *)(solPhaseNames + i*p_pname_len));
        
        pad = 0;
        for (c = 0; c < p_pname_len; c++) {  // pad with spaces
            if (pad == 1 || solPhaseNames[i*p_pname_len + c] == '\0' || solPhaseNames[i*p_pname_len + c] == ' ') {
                pad = 1;
                solPhaseNames[i*p_pname_len + c] = ' ';
            }
        }
        if (DBGL > 10)  {
            fprintf(stderr, "curCombination = %d, i = %d, solChoice = %d, string to = (%.14s), string from = (%s)\n", 
              curCombination, i, solChoice, (char *)(solPhaseNames + i*p_pname_len), &(allSolPhaseNames[irregArrPos(numAllSolPhaseNames, i, solChoice)*p_pname_len]));
        }
    }

    fread = fopen("datfiles/build_footer.dat", "r");
    if (fread == 0) {
        fprintf(stderr, "!! Error opening BUILD configuration file / output file\n");
        exit(1);
    }
    do {
        ch = fgetc(fread);
        if (ch != EOF) fputc(ch, fwrite);
        else break;
    } while(1);
    fclose(fwrite);
    fclose(fread);


    return 1;
}

int readNumArray(FILE *fp, double *arr, int num) {
    char *str1, *str2, *line;
    int nextline, i, n;

    line = (char *)calloc(MAX_LINE_LENGTH, sizeof(char));

    if (arr == NULL) {
        arr = (double *)calloc(num, sizeof(double));
    }
    
    nextline = 1;
    while(nextline) {
        if(feof(fp)) {
            fprintf(stderr, "!! Unexpected EOF\n");
            exit(1);
        }
        str1 = (char *)calloc(MAX_LINE_LENGTH, sizeof(char));
        fgets(line, MAX_LINE_LENGTH, fp);
        curline++;
        nextline = 0;
        strncpy(str1, line, MAX_LINE_LENGTH);
        for (i = 0; ; i++, safefree((void**)(&str1))) {
            str2 = strtok(str1, " \n");
            if (str2 == NULL) break;
            if (str2[0] == '#') {
                safefree((void**)(&str1));
                break;
            }
            if (i < num) {
                n = sscanf(str2, "%lg", &(arr[i]));
                if (n != 1) {
                    fprintf(stderr, "!! Non-number or other error while parsing, at line %d\n", curline);
                    return -1;
                }
            }
            if (DBGL > 10) fprintf(stderr, "!! Token: [%s]\n", str2);
        }
        if (i != num) {
            if (DBGL > 15) fprintf(stderr, "!! Read %d entries instead of expected %d (empty line?).\n", i, num);
            nextline = 1;
        }
    }
    
    if (DBGL > 5) {
        for (i = 0; i < num; i++) {
            fprintf(stderr, "Entry %d: %g\n", i, arr[i]);
        }
    }
    
    safefree((void **)(&line));
    
    return 0;
}

int getPhaseNumberPtr(char *solutionname) {
    char c[p_pname_len];
    
    strncpy(c, solutionname, p_pname_len-1);
    return getPhaseNumber(c);
}

int getPhaseNumber(char solutionname[p_pname_len]) {
    int i;
    
    if (DBGL > 5) fprintf(stderr, "!! Looking for <%.14s>...\n", solutionname);
    
    for (i = 0; i < num_ph; i++) {
        if(strncmp(solutionname, solPhaseNames + i*p_pname_len, p_pname_len-1) == 0) {
            if (DBGL > 5) fprintf(stderr, "!! Returning %d\n", i);
            return i;
        }
    }
    
    if (DBGL > 0) fprintf(stderr, "!! Phase (%s) not found in the list.\n", solutionname);
    return -1;
}

/* processes and stores the results from PerpleX in to result struct */
/* TODO: copy the used solution model combination here, too */
Result pToResult(double P, double T, int nphases, double *wtphases,
                 double *cphases, char *namephases, double *bulkcomp, 
                 int experiment, int id) {
    Result r;
    int i, j, k;
    
    r.P = P;
    r.T = T;
    r.id = id;
    r.exp = experiment;
    r.type = RES_CAL;
    r.bulkcomp = (double *)calloc(num_c, sizeof(double));
    r.phasewt = (double *)calloc(num_ph, sizeof(double));
    r.phasecomp = (double *)calloc(num_c*num_ph, sizeof(double));
    r.phasepresent = (int *)calloc(num_ph, sizeof(int));
    
    if (bulkcomp != NULL) {
        for (i = 0; i < num_c; i++) {
            r.bulkcomp[i] = bulkcomp[i];
        }
    } else {
        for (i = 0; i < num_c; i++) {
            r.bulkcomp[i] = VAL_UNKNOWN;
        }
    }

    for (i = 0; i < num_ph; i++) {
        r.phasewt[i] = 0.0;
        r.phasepresent[i] = 0.0;
        for (j = 0; j < num_c; j++) {
            r.phasecomp[i*num_c + j] = VAL_UNKNOWN;
        }
    }
    
    for (i = 0; i < nphases; i++) {
        k = getPhaseNumberPtr(namephases + i*p_pname_len);
        if (k >= 0) {
            r.phasewt[k] = wtphases[i] / 100.0;

            if (!isnormal(r.phasewt[k])) {
                fprintf(stderr, "!! Phase %d weight from PerpleX not normal (e.g. NaN). Exiting.\n", k);
                exit(1);
                r.phasewt[k] = VAL_UNKNOWN;
            }
            
            if (wtphases[i] > 0)
                r.phasepresent[i] = 1;
                
            for (j = 0; j < num_c; j++) {
                if (isnormal(cphases[i*num_c + j])) {
                    r.phasecomp[k*num_c + j] = cphases[i*num_c + j] / 100.0;
                } else {
                    //fprintf(stderr, "!! Phase %d component %d weight from PerpleX not normal (e.g. NaN). Exiting.\n", k, j);
                    //exit(1);
                    r.phasecomp[k*num_c + j] = VAL_UNKNOWN;
                }
            }
        } else {
            if (DBGL > 0) fprintf(stderr, "!! PerpleX returned a phase not in genPhaseNames list\n");
        }
    }
    
    return r;
}

int isknown(double val) {
    if (val - 1.0 < VAL_UNKNOWN && val + 1.0 > VAL_UNKNOWN) {
        return 0;
    } else {
        return 1;
    }
}

double evalExperiment(Experiments *meas, Experiments *calc) {
    int i, j, skipped_ph, skipped_res;
    double res_err = 0;
    double tot_err = 0;
    double diff;
    int res_count = 0;

    if (meas->count != calc->count) {
        fprintf(stderr, "!! Result count of measured (%d) and calculated (%d) experiment do not match\n", meas->count, calc->count);
        exit(1);
    }

    skipped_res = 0;
    for (i = 0; i < meas->count; i++) {
        res_err = 0;
        skipped_ph = 0;
        if (meas->result[i].P != calc->result[i].P ||
                meas->result[i].T != meas->result[i].T) {
            fprintf(stderr, "!! Calc result %d not in same P/T as measured result. Skipping.\n", i);
            skipped_res++;
            res_err = 0;
            continue;
        }
        res_count++;
        if (meas->result[i].flag_ph_ab == 1) {  // phase abundancies are given in wt%
            for (j = 0; j < num_ph; j++) {
                if (isknown(meas->result[i].phasewt[j]) && isknown(calc->result[i].phasewt[i])) {
                    diff = meas->result[i].phasewt[j] - calc->result[i].phasewt[j];
                    res_err += diff*diff;
                } else {
                    if (DBGL > 15) {
                        fprintf(stderr, "!! Skipping: phase %d, result %d. Unknown value.\n", j, i);
                    }
                    skipped_ph++;
                }
            }
            if (num_ph == skipped_ph) {
                skipped_res++;
                res_err = 0;
                if (DBGL > 5) fprintf(stderr, "!! Skipping result %d. Unknown value.\n", i);
            } else {
                res_err = sqrt(res_err) / (num_ph - skipped_ph);
            }
            fprintf(stderr, " - Result %d, res_err = %g\n", i, res_err);
            tot_err += res_err;
        } else if (meas->result[i].flag_ph_ab == 2) { // phase abundancies are given as exists/non-existing
            for (j = 0; j < num_ph; j++) {
                if (isknown(meas->result[i].phasewt[j]) && isknown(calc->result[i].phasewt[i])) {
                    if (meas->result[i].phasewt[j] > 0) {
                        if (calc->result[i].phasewt[j] > 0) {
                            res_err += 0;
                        } else {
                            res_err += 0.5;
                        }
                    } else if (meas->result[i].phasewt[j] == 0) {
                        if (calc->result[i].phasewt[j] == 0) {
                            res_err += 0;
                        } else {
                            res_err += 0.5;
                            //res_err += 1/(1-calc->result[i].phasewt[j]) - 1;
                        }
                    } else {
                        fprintf(stderr, "!! Invalid phasewt value\n");
                        exit(1);
                    }
                } else {
                    skipped_ph++;
                }
                if (num_ph == skipped_ph) {
                    skipped_res++;
                    res_err = 0;
                } else {
                    res_err = res_err / (num_ph - skipped_ph);
                }
            }         
            tot_err += res_err;
        } else {
            fprintf(stderr, "!! Invalid flag_ph_ab flag %d\n", meas->result[i].flag_ph_ab);
            exit(1);
        }
    }
    if (skipped_res == res_count) {
        fprintf(stderr, "!! Skipped results == result count. Cannot evaluate. Exiting.\n");
        exit(1);
    } else {
        tot_err = tot_err / (res_count - skipped_res);
    }
    fprintf(stderr, " * Experiment: tot_err = %g\n\n\n\n", tot_err);

    return tot_err;
}

int compareLists(char *list1, char *list2, int n, int len) {
    int ret = 0;
    int i;
    for (i = 0; i < n; i++) {
        if(strncmp(list1 + n*len, list2 + n*len, len) != 0) {
            ret += 1;
            break;
        }
    }
    return ret;
}


int main(int argc, char **argv) {
    int i, j, ires, iexp;
    double expErr;
    
    int nphases;    // number of stable phases returned by PerpleX, != NUM_PH
    double *wtphases = 0;   // weights of phases ...
    double *cphases = 0;    // ... and compositions of phases returned by PerpleX
    
    char *namephases = 0;

    double use_P, use_T;
    
    initialize();

    for (iexp = 0; iexp < num_var; iexp++) {
        if (DYNAMIC_MEM_ALLOC) {
            if (DBGL > 5) fprintf(stderr, "Reserving %ld kilobytes mem for calc result data array...\n", calc.experiments[iexp].count*sizeof(Result));
            calc.experiments[iexp].result = (Result *)calloc(calc.experiments[iexp].count, sizeof(Result));
        }

        if (!nextSolutionModelCombination()) {
            fprintf(stderr, "!! Error while trying to apply the next solution model combination (num %d).\n", iexp);
            exit(1);
        }
        ini_phaseq("datfiles/errormin");

        if (meas.experiments[0].count != calc.experiments[iexp].count) {
            fprintf(stderr, "!! Experiment %d: num of results different in meas and calc.\n", iexp);
        }

        for (ires = 0; ires < meas.experiments[0].count; ires++) {
            if (compareLists(solPhaseNames, solPhaseNames, num_ph, p_pname_len-1)) {    // todo: compare prev and this solution model list, if changes, call ini_phaseq
                fprintf(stderr, "!! Adjusting solution models ...\n");
                ini_phaseq("datfiles/errormin");
            }
    
            phaseq(meas.experiments[0].result[ires].P, meas.experiments[0].result[ires].T, num_c, 
                   meas.experiments[0].result[ires].bulkcomp, &nphases, &wtphases, &cphases,
                   &namephases, 0);
            calc_res_id++;
            
            use_P = meas.experiments[0].result[ires].P;
            use_T = meas.experiments[0].result[ires].T;
            calc.experiments[iexp].result[ires] = pToResult(use_P, use_T, nphases, wtphases, cphases,  
                                                            namephases, meas.experiments[0].result[ires].bulkcomp,
                                                            iexp, calc_res_id);
    
            /*
            // print it
            fprintf(stderr, "\n\n\n              \twt%%\t");
            for (i = 0; i < num_c; i++) {
                fprintf(stderr, "%.5s\t", compNames + i*C_NAME_LEN);
            }
            fprintf(stderr, "\n");
            for (i = 0; i < num_ph; i++) {
                fprintf(stderr, "%.14s:\t", genPhaseNames + i*p_pname_len);
                fprintf(stderr, "%.4g\t", calc.experiments[iexp].result[ires].phasewt[i]);
                for (j = 0; j < num_c; j++) {
                    fprintf(stderr, "%.4g\t", calc.experiments[iexp].result[ires].phasecomp[i*num_c + j]);
                }
                fprintf(stderr, "\n");
            }
            */

            fprintf(stderr, "** Exp %d, result %d (id %d):\n", iexp, ires, calc_res_id);
            fprintf(stderr, "Phases:\n");
            fprintf(stderr, "Phase         \tCalc\tMeas\n");
            for (i = 0; i < num_ph; i++) {
                fprintf(stderr, "%.14s\t%g\t%g\n", genPhaseNames + i*p_pname_len, calc.experiments[iexp].result[ires].phasewt[i], meas.experiments[0].result[ires].phasewt[i]);
            }
            fprintf(stderr, "\n\n");

            safefree((void **)(&cphases));
            safefree((void **)(&wtphases));
            safefree((void **)(&namephases));

        }

        expErr = evalExperiment(&(meas.experiments[0]), &(calc.experiments[iexp]));
        fprintf(fp_results, "%d %g\n", iexp, expErr);

        if (DYNAMIC_MEM_ALLOC) {
            for (j = 0; j < calc.experiments[iexp].count; j++) {
                safefree((void **)(&(calc.experiments[iexp].result[j].bulkcomp)));
                safefree((void **)(&(calc.experiments[iexp].result[j].phasewt)));
                safefree((void **)(&(calc.experiments[iexp].result[j].phasecomp)));
                safefree((void **)(&(calc.experiments[iexp].result[j].phasepresent)));
            }
            safefree((void **)(&(calc.experiments[iexp].result)));
        }
    }

    fclose(fp_results);
    
    return 0;
}

